# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################



import datetime
import os
from xml.etree.ElementTree import Element, ElementTree, SubElement

import sickrage
from sickrage.core.common import dateFormat
from sickrage.core.helpers import replace_extension, indent_xml, chmod_as_parent
from sickrage.metadata_providers.mediabrowser import MediaBrowserMetadata
from sickrage.series_providers.exceptions import SeriesProviderEpisodeNotFound, SeriesProviderSeasonNotFound


class Mede8erMetadata(MediaBrowserMetadata):
    """
    Metadata generation class for Mede8er based on the MediaBrowser.

    The following file structure is used:

    show_root/series.xml                    (show metadata)
    show_root/folder.jpg                    (poster)
    show_root/fanart.jpg                    (fanart)
    show_root/Season ##/folder.jpg          (season thumb)
    show_root/Season ##/filename.ext        (*)
    show_root/Season ##/filename.xml        (episode metadata)
    show_root/Season ##/filename.jpg        (episode thumb)
    """

    def __init__(self,
                 show_metadata=False,
                 episode_metadata=False,
                 fanart=False,
                 poster=False,
                 banner=False,
                 episode_thumbnails=False,
                 season_posters=False,
                 season_banners=False,
                 season_all_poster=False,
                 season_all_banner=False):

        MediaBrowserMetadata.__init__(
            self, show_metadata, episode_metadata, fanart,
            poster, banner, episode_thumbnails, season_posters,
            season_banners, season_all_poster, season_all_banner
        )

        self.name = "Mede8er"

        self.fanart_name = "fanart.jpg"

        # web-ui metadata template
        # self.eg_show_metadata = "series.xml"
        self.eg_episode_metadata = "Season##\\<i>filename</i>.xml"
        self.eg_fanart = "fanart.jpg"
        # self.eg_poster = "folder.jpg"
        # self.eg_banner = "banner.jpg"
        self.eg_episode_thumbnails = "Season##\\<i>filename</i>.jpg"
        # self.eg_season_posters = "Season##\\folder.jpg"
        # self.eg_season_banners = "Season##\\banner.jpg"
        # self.eg_season_all_poster = "<i>not supported</i>"
        # self.eg_season_all_banner = "<i>not supported</i>"

    def get_episode_file_path(self, ep_obj):
        return replace_extension(ep_obj.location, self._ep_nfo_extension)

    @staticmethod
    def get_episode_thumb_path(ep_obj):
        return replace_extension(ep_obj.location, 'jpg')

    def _show_data(self, show_obj):
        """
        Creates an elementTree XML structure for a MediaBrowser-style series.xml
        returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """

        root_node = Element("details")
        tv_node = SubElement(root_node, "movie")
        tv_node.attrib["isExtra"] = "false"
        tv_node.attrib["isSet"] = "false"
        tv_node.attrib["isTV"] = "true"

        series_provider_language = show_obj.lang or sickrage.app.config.general.series_provider_default_language
        series_info = show_obj.series_provider.get_series_info(show_obj.series_id, language=series_provider_language)
        if not series_info:
            return False

        # check for title and id
        if not (getattr(series_info, 'name', None) and getattr(series_info, 'id', None)):
            sickrage.app.log.info("Incomplete info for "
                                  "show with id " + str(show_obj.series_id) + " on " + show_obj.series_provider.name + ", skipping it")
            return False

        SeriesName = SubElement(tv_node, "title")
        SeriesName.text = series_info['name']

        if getattr(series_info, "genre", None):
            Genres = SubElement(tv_node, "genres")
            for genre in series_info['genre']:
                cur_genre = SubElement(Genres, "Genre")
                cur_genre.text = genre['name'].strip()

        if getattr(series_info, 'firstAired', None):
            FirstAired = SubElement(tv_node, "premiered")
            FirstAired.text = series_info['firstAired']

        if getattr(series_info, "firstAired", None):
            try:
                year_text = str(datetime.datetime.strptime(series_info["firstAired"], dateFormat).year)
                if year_text:
                    year = SubElement(tv_node, "year")
                    year.text = year_text
            except Exception:
                pass

        if getattr(series_info, 'overview', None):
            plot = SubElement(tv_node, "plot")
            plot.text = series_info["overview"]

        if getattr(series_info, 'rating', None):
            try:
                rating = int(float(series_info['rating']) * 10)
            except ValueError:
                rating = 0

            if rating:
                Rating = SubElement(tv_node, "rating")
                Rating.text = str(rating)

        if getattr(series_info, 'status', None):
            Status = SubElement(tv_node, "status")
            Status.text = series_info['status']

        # if getattr(series_info, "contentrating", None):
        #     mpaa = SubElement(tv_node, "mpaa")
        #     mpaa.text = series_info["contentrating"]

        if getattr(series_info, 'imdbId', None):
            imdb_id = SubElement(tv_node, "id")
            imdb_id.attrib["moviedb"] = "imdb"
            imdb_id.text = series_info['imdbId']

        if getattr(series_info, 'id', None):
            series_id = SubElement(tv_node, "series_id")
            series_id.text = str(series_info['id'])

        if getattr(series_info, 'runtime', None):
            Runtime = SubElement(tv_node, "runtime")
            Runtime.text = str(series_info['runtime'])

        cast = SubElement(tv_node, "cast")
        for person in series_info['people']:
            if 'name' not in person or not person['name'].strip():
                continue

            if person['role'].strip() == 'Actor':
                cur_actor = SubElement(cast, "actor")

                cur_actor_role = SubElement(cur_actor, "role")
                cur_actor_role.text = person['role'].strip()

                cur_actor_name = SubElement(cur_actor, "name")
                cur_actor_name.text = person['name'].strip()

                if person['imageUrl'].strip():
                    cur_actor_thumb = SubElement(cur_actor, "thumb")
                    cur_actor_thumb.text = person['imageUrl'].strip()

        indent_xml(root_node)

        data = ElementTree(root_node)

        return data

    def _ep_data(self, ep_obj):
        """
        Creates an elementTree XML structure for a MediaBrowser style episode.xml
        and returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """

        eps_to_write = [ep_obj] + ep_obj.related_episodes

        series_provider_language = ep_obj.show.lang or sickrage.app.config.general.series_provider_default_language
        series_info = ep_obj.show.series_provider.get_series_info(ep_obj.show.series_id, language=series_provider_language)
        if not series_info:
            return False

        rootNode = Element("details")
        movie = SubElement(rootNode, "movie")

        movie.attrib["isExtra"] = "false"
        movie.attrib["isSet"] = "false"
        movie.attrib["isTV"] = "true"

        # write an MediaBrowser XML containing info for all matching episodes
        for curEpToWrite in eps_to_write:
            try:
                series_episode_info = series_info[curEpToWrite.season][curEpToWrite.episode]
            except (SeriesProviderEpisodeNotFound, SeriesProviderSeasonNotFound):
                sickrage.app.log.info(
                    f"Unable to find episode {curEpToWrite.season:d}x{curEpToWrite.episode:d} on {ep_obj.show.series_provider.name}"
                    f"... has it been removed? Should I delete from db?")
                return None

            if curEpToWrite == ep_obj:
                # root (or single) episode

                # default to today's date for specials if firstaired is not set
                if curEpToWrite.season == 0 and not getattr(series_episode_info, 'firstAired', None):
                    series_episode_info['firstAired'] = str(datetime.date.min)

                if not (getattr(series_episode_info, 'name', None) and getattr(series_episode_info, 'firstAired', None)):
                    return None

                episode = movie

                if curEpToWrite.name:
                    EpisodeName = SubElement(episode, "title")
                    EpisodeName.text = curEpToWrite.name

                SeasonNumber = SubElement(episode, "season")
                SeasonNumber.text = str(curEpToWrite.season)

                EpisodeNumber = SubElement(episode, "episode")
                EpisodeNumber.text = str(curEpToWrite.episode)

                if getattr(series_info, "firstAired", None):
                    try:
                        year_text = str(datetime.datetime.strptime(series_info["firstAired"], dateFormat).year)
                        if year_text:
                            year = SubElement(episode, "year")
                            year.text = year_text
                    except:
                        pass

                if getattr(series_info, "overview", None):
                    plot = SubElement(episode, "plot")
                    plot.text = series_info["overview"]

                if curEpToWrite.description:
                    Overview = SubElement(episode, "episodeplot")
                    Overview.text = curEpToWrite.description

                if getattr(series_info, 'contentrating', None):
                    mpaa = SubElement(episode, "mpaa")
                    mpaa.text = series_info["contentrating"]

                if not ep_obj.related_episodes and getattr(series_episode_info, "rating", None):
                    try:
                        rating = int((float(series_episode_info['rating']) * 10))
                    except ValueError:
                        rating = 0

                    if rating:
                        Rating = SubElement(episode, "rating")
                        Rating.text = str(rating)

                cast = SubElement(episode, "cast")
                for person in series_info['people']:
                    if 'name' not in person or not person['name'].strip():
                        continue

                    if person['role'].strip() == 'Actor':
                        cur_actor = SubElement(cast, "actor")
                        cur_actor_role = SubElement(cur_actor, "role")
                        cur_actor_role.text = person['role'].strip()

                        cur_actor_name = SubElement(cur_actor, "name")
                        cur_actor_name.text = person['name'].strip()

                        if person['imageUrl'].strip():
                            cur_actor_thumb = SubElement(cur_actor, "thumb")
                            cur_actor_thumb.text = person['imageUrl'].strip()
                    elif person['role'].strip() == 'Writer':
                        writer = SubElement(episode, "credits")
                        writer.text = series_episode_info['writer'].strip()
                    elif person['role'].strip() == 'Director':
                        director = SubElement(episode, "director")
                        director.text = series_episode_info['director'].strip()
                    elif person['role'].strip() == 'Guest Star':
                        cur_actor = SubElement(cast, "actor")
                        cur_actor_name = SubElement(cur_actor, "name")
                        cur_actor_name.text = person['name'].strip()

            else:
                episode = movie

                # append data from (if any) related episodes
                if curEpToWrite.name:
                    EpisodeName = SubElement(episode, "title")

                    if not EpisodeName.text:
                        EpisodeName.text = curEpToWrite.name
                    else:
                        EpisodeName.text = EpisodeName.text + ", " + curEpToWrite.name

                if curEpToWrite.description:
                    Overview = SubElement(episode, "episodeplot")

                    if not Overview.text:
                        Overview.text = curEpToWrite.description
                    else:
                        Overview.text = Overview.text + "\r" + curEpToWrite.description

        indent_xml(rootNode)
        data = ElementTree(rootNode)

        return data

    def write_show_file(self, show_obj):
        """
        Generates and writes show_obj's metadata under the given path to the
        filename given by get_show_file_path()

        show_obj: TVShow object for which to create the metadata

        path: An absolute or relative path where we should put the file. Note that
                the file name will be the default show_file_name.

        Note that this method expects that _show_data will return an ElementTree
        object. If your _show_data returns data in another format yo'll need to
        override this method.
        """

        data = self._show_data(show_obj)
        if not data:
            return False

        nfo_file_path = self.get_show_file_path(show_obj)
        nfo_file_dir = os.path.dirname(nfo_file_path)

        try:
            if not os.path.isdir(nfo_file_dir):
                sickrage.app.log.debug("Metadata dir didn't exist, creating it at " + nfo_file_dir)
                os.makedirs(nfo_file_dir)
                chmod_as_parent(nfo_file_dir)

            sickrage.app.log.debug("Writing show nfo file to " + nfo_file_path)

            nfo_file = open(nfo_file_path, 'wb')

            data.write(nfo_file)
            nfo_file.close()
            chmod_as_parent(nfo_file_path)
        except IOError as e:
            sickrage.app.log.error(
                "Unable to write file to " + nfo_file_path + " - are you sure the folder is writable? {}".format(e))
            return False

        return True

    def write_ep_file(self, ep_obj):
        """
        Generates and writes ep_obj's metadata under the given path with the
        given filename root. Uses the episode's name with the extension in
        _ep_nfo_extension.

        ep_obj: TVEpisode object for which to create the metadata

        file_name_path: The file name to use for this metadata. Note that the extension
                will be automatically added based on _ep_nfo_extension. This should
                include an absolute path.

        Note that this method expects that _ep_data will return an ElementTree
        object. If your _ep_data returns data in another format yo'll need to
        override this method.
        """

        data = self._ep_data(ep_obj)

        if not data:
            return False

        nfo_file_path = self.get_episode_file_path(ep_obj)
        nfo_file_dir = os.path.dirname(nfo_file_path)

        try:
            if not os.path.isdir(nfo_file_dir):
                sickrage.app.log.debug("Metadata dir didn't exist, creating it at " + nfo_file_dir)
                os.makedirs(nfo_file_dir)
                chmod_as_parent(nfo_file_dir)

            sickrage.app.log.debug("Writing episode nfo file to " + nfo_file_path)

            nfo_file = open(nfo_file_path, 'wb')

            data.write(nfo_file)
            nfo_file.close()
            chmod_as_parent(nfo_file_path)
        except IOError as e:
            sickrage.app.log.warning(
                "Unable to write file to " + nfo_file_path + " - are you sure the folder is writable? {}".format(e))
            return False

        return True



"""Initial migration

Revision ID: 4
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.

revision = '4'
down_revision = '3'


def upgrade():
    op.add_column('general', sa.Column('sso_api_key', sa.Text, default=''))


def downgrade():
    pass



# Author: seedboy
# URL: https://github.com/seedboy
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import re
from urllib.parse import urljoin

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import bs4_parser, convert_size, validate_url
from sickrage.search_providers import TorrentProvider


class IPTorrentsProvider(TorrentProvider):
    def __init__(self):
        super(IPTorrentsProvider, self).__init__("IPTorrents", 'https://iptorrents.eu', True)

        self.enable_cookies = True
        self.required_cookies = ('uid', 'pass')

        # custom settings
        self.custom_settings = {
            'username': '',
            'password': '',
            'custom_url': '',
            'freeleech': False,
            'minseed': 0,
            'minleech': 0
        }

        self.categories = '73=&60='

        self.cache = TVCache(self, min_time=10)

    @property
    def urls(self):
        return {
            'login': f'{self.url}/torrents',
            'search': f'{self.url}/t?%s%s&q=%s&qf=#torrents'
        }

    def login(self):
        return self.cookie_login('sign in')

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        results = []

        if not self.login():
            return results

        freeleech = '&free=on' if self.custom_settings['freeleech'] else ''

        for mode in search_strings:
            sickrage.app.log.debug("Search Mode: %s" % mode)
            for search_string in search_strings[mode]:
                if mode != 'RSS':
                    sickrage.app.log.debug("Search string: %s " % search_string)

                # URL with 50 tv-show results, or max 150 if adjusted in IPTorrents profile
                search_url = self.urls['search'] % (self.categories, freeleech, search_string)
                search_url += ';o=seeders' if mode != 'RSS' else ''

                if self.custom_settings['custom_url']:
                    if not validate_url(self.custom_settings['custom_url']):
                        sickrage.app.log.warning("Invalid custom url: {}".format(self.custom_settings['custom_url']))
                        return results

                    search_url = urljoin(self.custom_settings['custom_url'], search_url.split(self.url)[1])

                resp = self.session.get(search_url)
                if not resp or not resp.text:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                results += self.parse(resp.text, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        data = re.sub(r'(?im)<button.+?<[/]button>', '', data, 0)

        with bs4_parser(data) as html:
            torrent_table = html.find('table', id='torrents')
            torrents = torrent_table('tr') if torrent_table else []

            # Continue only if one Release is found
            if len(torrents) < 2 or html.find(text='No Torrents Found!'):
                sickrage.app.log.debug("Data returned from provider does not contain any torrents")
                return results

            for torrent in torrents[1:]:
                try:
                    title = torrent('td')[1].find('a').text
                    download_url = self.url + torrent('td')[3].find('a')['href']
                    if not all([title, download_url]):
                        continue

                    size = convert_size(torrent('td')[5].text, -1)
                    seeders = int(torrent('td')[7].contents[0])
                    leechers = int(torrent('td')[8].contents[0])

                    results += [{
                        'title': title,
                        'link': download_url,
                        'size': size,
                        'seeders': seeders,
                        'leechers': leechers
                    }]

                    if mode != 'RSS':
                        sickrage.app.log.debug("Found result: {}".format(title))
                except Exception:
                    sickrage.app.log.error("Failed parsing provider")

        return results



import re
from urllib.parse import urljoin

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import try_int, convert_size, validate_url, bs4_parser
from sickrage.search_providers import TorrentProvider


class LeetxProvider(TorrentProvider):
    def __init__(self):
        super().__init__("1337x", "https://1337x.to", False)

        # custom settings
        self.custom_settings = {
            'custom_url': '',
            'minseed': 0,
            'minleech': 0
        }

        self.cache = TVCache(self)

    @property
    def urls(self):
        return {
            "search": f'{self.url}/sort-search/%s/seeders/desc/',
            'rss': f'{self.url}/cat/TV/'
        }

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        results = []

        for mode in search_strings:
            sickrage.app.log.debug("Search Mode: {mode}".format(mode=mode))
            for search_string in {*search_strings[mode]}:
                if mode != "RSS":
                    sickrage.app.log.debug("Search String: {search_string}".format(search_string=search_string))
                    search_url = self.urls['search'] % search_string
                else:
                    search_url = self.urls["rss"]

                if self.custom_settings['custom_url']:
                    if not validate_url(self.custom_settings['custom_url']):
                        sickrage.app.log.warning("Invalid custom url: {0}".format(self.custom_settings['custom_url']))
                        return results
                    search_url = urljoin(self.custom_settings['custom_url'], search_url.split(self.url)[1])

                page = 0
                while page <= 10:
                    page += 1
                    resp = self.session.get(urljoin(search_url, f'{page}/'))
                    if not resp or not resp.text:
                        sickrage.app.log.debug("No data returned from provider")
                        continue

                    results += self.parse(resp.text, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        with bs4_parser(data) as html:
            torrent_table = html.find('table', {'class': 'table-list'})
            if not torrent_table:
                sickrage.app.log.debug('Data returned from provider does not contain any torrents')
                return results

            rows = torrent_table.tbody.find_all('tr')
            for row in rows:
                try:
                    name_col = row.find('td', {'class': 'coll-1 name'})
                    title = name_col.a.next_sibling.get_text(strip=True)
                    magnet = self.extract_magnet_link(self.url + name_col.a.next_sibling['href'])
                    seeders = try_int(row.find('td', {'class': 'coll-2 seeds'}).get_text(strip=True))
                    leechers = try_int(row.find('td', {'class': 'coll-3 leeches'}).get_text(strip=True))
                    size = convert_size(row.find('td', {'class': 'coll-4'}).get_text(strip=True), -1)

                    if not all([title, magnet]):
                        continue

                    results += [
                        {"title": title, "link": magnet, "size": size, "seeders": seeders, "leechers": leechers}
                    ]

                    if mode != "RSS":
                        sickrage.app.log.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))
                except Exception as error:
                    sickrage.app.log.debug(f"Failed parsing provider. Traceback: {error}")
                    continue

        return results

    def extract_magnet_link(self, url):
        resp = self.session.get(url)
        with bs4_parser(resp.text) as html:
            magnet_link = html.find('a', href=re.compile(r'^magnet:\?')).get('href')
            return magnet_link



"""Initial migration

Revision ID: 10
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import json
import os
from json import JSONDecodeError

import sqlalchemy as sa
from alembic import op
from keycloak.exceptions import KeycloakClientError
from sqlalchemy import orm, inspect

import sickrage

# revision identifiers, used by Alembic.
from sickrage.core import ConfigDB

revision = '10'
down_revision = '9'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    oauth2_token = sa.Table('oauth2_token', meta, autoload=True)

    refresh_token = None
    with op.get_context().begin_transaction():
        for row in conn.execute(oauth2_token.select()):
            refresh_token = row.refresh_token
            if refresh_token:
                break

    try:
        if refresh_token:
            certs = sickrage.app.auth_server.certs()
            if certs:
                new_token = sickrage.app.auth_server.refresh_token(refresh_token)
                if new_token:
                    decoded_token = sickrage.app.auth_server.decode_token(new_token['access_token'], certs)
                    apikey = decoded_token.get('apikey')
                    if apikey:
                        session = sickrage.app.config.db.session()
                        general = session.query(ConfigDB.General).one()
                        general.sso_api_key = apikey
                        session.commit()
    except (KeycloakClientError, orm.exc.NoResultFound):
        pass

    if inspect(conn).has_table('oauth2_token'):
        op.drop_table('oauth2_token')


def downgrade():
    # Operations to reverse the above upgrade go here.
    pass



"""Initial migration

Revision ID: 5
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '5'
down_revision = '4'


def upgrade():
    op.create_table(
        'announcements',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('hash', sa.String(255), unique=True, nullable=False),
        sa.Column('seen', sa.Boolean)
    )


def downgrade():
    pass



"""Initial migration

Revision ID: 6
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import babelfish
from alembic import op

# revision identifiers, used by Alembic.

revision = '6'
down_revision = '5'


def upgrade():
    pass


def downgrade():
    pass



import re
import traceback
import urllib
from collections import OrderedDict
from urllib.parse import urljoin

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import bs4_parser, try_int, convert_size, validate_url
from sickrage.core.tv.show.helpers import find_show
from sickrage.search_providers import TorrentProvider


class KickAssTorrentsProvider(TorrentProvider):
    def __init__(self):
        super(KickAssTorrentsProvider, self).__init__('KickAssTorrents', 'https://kickasskat.org', False)

        # custom settings
        self.custom_settings = {
            'custom_url': '',
            'confirmed': False,
            'minseed': 0,
            'minleech': 0
        }

        self.mirrors = []
        self.disabled_mirrors = []

        # https://kickasskat.org/tv?field=time_add&sorder=desc
        # https://kickasskat.org/usearch/{query} category:tv/?field=seeders&sorder=desc

        self.cache = TVCache(self)

        self.rows_selector = dict(class_=re.compile(r"even|odd"), id=re.compile(r"torrent_.*_torrents"))

    @property
    def urls(self):
        return {
            "search": f'{self.url}/usearch/%s/',
            "rss": f'{self.url}/tv/'
        }

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        results = []
        if not (self.url and self.urls):
            self.find_domain()
            if not (self.url and self.urls):
                return results

        show_object = find_show(series_id, series_provider_id)
        if not show_object:
            return results

        search_params = OrderedDict(field="seeders", sorder="desc")

        for mode in search_strings:
            sickrage.app.log.debug("Search Mode: {mode}".format(mode=mode))
            for search_string in {*search_strings[mode]}:
                # search_params["q"] = (search_string, None)[mode == "RSS"]
                search_params["field"] = ("seeders", "time_add")[mode == "RSS"]

                if mode != "RSS":
                    if show_object.anime:
                        continue

                    sickrage.app.log.debug(f"Search String: {search_string}")

                    search_url = self.urls['search'] % search_string + f' category:{("tv", "anime")[show_object.anime]}'
                else:
                    search_url = self.urls["rss"]

                if self.custom_settings['custom_url']:
                    if not validate_url(self.custom_settings['custom_url']):
                        sickrage.app.log.warning("Invalid custom url: {0}".format(self.custom_settings['custom_url']))
                        return results
                    search_url = urljoin(self.custom_settings['custom_url'], search_url.split(self.url)[1])

                resp = self.session.get(search_url, params=search_params, random_ua=True)
                if not resp or not resp.text:
                    sickrage.app.log.info("{url} did not return any data, it may be disabled. Trying to get a new domain".format(url=self.url))
                    self.disabled_mirrors.append(self.url)
                    self.find_domain()
                    if self.url in self.disabled_mirrors:
                        sickrage.app.log.info("Could not find a better mirror to try.")
                        sickrage.app.log.info("The search did not return data, if the results are on the site maybe try a custom url, or a different one")
                        return results

                    # This will recurse a few times until all of the mirrors are exhausted if none of them work.
                    return self.search(search_strings, age, series_id, series_provider_id, season, episode)

                results += self.parse(resp.text, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        with bs4_parser(data) as html:
            labels = [cell.get_text() for cell in html.find(class_="firstr")("th")]
            sickrage.app.log.info("Found {} results".format(len(html("tr", **self.rows_selector))))
            for result in html("tr", **self.rows_selector):
                try:
                    download_url = urllib.parse.unquote_plus(result.find(title="Torrent magnet link")["href"].split("url=")[1])
                    parsed_magnet = urllib.parse.parse_qs(download_url)
                    title = result.find(class_="torrentname").find(class_="cellMainLink").get_text(strip=True)
                    if title.endswith("..."):
                        title = parsed_magnet["dn"][0]

                    if not (title and download_url):
                        if mode != "RSS":
                            sickrage.app.log.debug("Discarding torrent because We could not parse the title and url")
                        continue

                    seeders = try_int(result.find(class_="green").get_text(strip=True))
                    leechers = try_int(result.find(class_="red").get_text(strip=True))

                    if self.custom_settings['confirmed'] and not result.find(class_="ka-green"):
                        if mode != "RSS":
                            sickrage.app.log.debug("Found result " + title + " but that doesn't seem like a verified result so I'm ignoring it")
                        continue

                    torrent_size = result("td")[labels.index("size")].get_text(strip=True)
                    size = convert_size(torrent_size, -1)

                    results += [
                        {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers}
                    ]

                    if mode != "RSS":
                        sickrage.app.log.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))
                except (AttributeError, TypeError, KeyError, ValueError, Exception):
                    sickrage.app.log.info(traceback.format_exc())
                    continue

        return results

    def find_domain(self):
        resp = self.session.get("https://ww1.kickass.help/")
        if not resp or not resp.text:
            return self.url

        with bs4_parser(resp.text) as html:
            mirrors = html(class_="domainLink")
            if mirrors:
                self.mirrors = []

            for mirror in mirrors:
                domain = mirror["href"].rstrip('/')
                if domain and domain not in self.disabled_mirrors:
                    self.mirrors.append(domain)

        if self.mirrors:
            self.url = self.mirrors[0]
            sickrage.app.log.info("Setting mirror to use to {url}".format(url=self.url))
        else:
            sickrage.app.log.warning(
                "Unable to get a working mirror for KickassTorrent. You might need to enable another provider and disable KAT until it starts working again."
            )

        return self.url



"""Initial migration

Revision ID: 4
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import babelfish
from alembic import op

# revision identifiers, used by Alembic.

revision = '5'
down_revision = '4'


def upgrade():
    conn = op.get_bind()
    row = conn.execute(f"SELECT series_provider_default_language FROM general").first()
    if len(row.series_provider_default_language) == 2:
        lang = babelfish.Language.fromalpha2(row.series_provider_default_language)
        conn.execute(f'UPDATE general SET series_provider_default_language = "{lang.alpha3}"')


def downgrade():
    pass
